// -*-C++-*-
//
//  This file is part of SSim, a simple discrete-event simulator.
//  http://www.cs.colorado.edu/serl/ssim/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2003 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: speedtest.cc,v 1.4 2005/01/20 17:26:05 carzanig Exp $
//
#include <cstring>
#include <string>
#include <iostream>

#include <siena/ssim.h>
#include <siena/tprocess.h>

using namespace ssim;
using std::string;
using std::cerr;
using std::cout;
using std::endl;

static bool do_print_trace;

//
// this is a simple iterative process
//
class TClient: public TProcess {
protected:
    int iterations;
    int interval;
    string name;
 public:
    TClient(const string &n, int i, int delta): 
	name(n), iterations(i), interval(delta)  {};

    virtual void		main(void);
};

class QTClient: public TClient {
 public:
    QTClient(const string &n, int i, int delta): TClient(n, i, delta) {};

    virtual void		main(void);
};

void TClient::main() { 
    for(int i = 0; i < iterations; ++i) {
    cout << "I am " << name << ", the time is " << Sim::clock();
    cout << ", " << iterations - i << " to go!" << endl;
	Sim::self_signal_event(NULL, interval);
	wait_for_event();
    }
    cout << "I am " << name << ", the time is " << Sim::clock();
    cout << ", no more iteration.  Bye!" << endl;
}

void QTClient::main() { 
    for(int i = 0; i < iterations; ++i) {
	Sim::self_signal_event(NULL, interval);
	wait_for_event();
    }
}

//
// this is a simple iterative process
//
class Client: public Process {
protected:
    int iterations;
    int interval;
    int duration;
    string name;
 public:
    Client(const string &n, int i, int delta): 
	name(n), iterations(i), interval(delta)  {};

    virtual void init(void);
    virtual void process_event(const Event *);
};

//
// this is a simple iterative process
//
class QClient: public Client {
 public:
    QClient(const string &n, int i, int delta): Client(n, i, delta) {};

    virtual void process_event(const Event *);
};

void Client::init() { 
    //
    // we start immediately by giving ourselves an immediate "timeout"
    //
    Sim::self_signal_event(NULL, 0);
}

void Client::process_event(const Event * e) { 
    if (e != 0) {
	cerr << "Client: sorry, I have not been programmed to handle events"
	     << "only \"timeouts\", please." << endl;
	return;
    }
    cout << "I am " << name << ", the time is " << Sim::clock();
    if (iterations > 0) {
	cout << ", " << iterations << " to go!";
	--iterations;
	Sim::self_signal_event(NULL, interval);
    } else {
	cout << ", no more iteration.  Bye!";
    }
    cout << endl;
}

void QClient::process_event(const Event * e) { 
    if (e != 0) {
	cerr << "Client: sorry, I have not been programmed to handle events"
	     << "only \"timeouts\", please." << endl;
	return;
    }
    if (iterations > 0) {
	--iterations;
	Sim::self_signal_event(NULL, interval);
    }
}

void print_usage(const char * progname) {
  cerr << "Usage: " << progname << " [-t] [-q] [-N <iterations>]" << endl;
}

int main(int argc, char *argv[]) {
    std::ios_base::sync_with_stdio(false);
    Process * c1, * c2, * c3, * c4;
    bool threaded_version = false;
    bool quiet_version = false;
    unsigned long iterations = 250000;  // total = 1M iterations

    for(int i = 1; i < argc; ++i) {
	if (strcmp(argv[i], "-t") == 0) {
	    threaded_version = true;
	} else if (strcmp(argv[i], "-q") == 0) {
	    quiet_version = true;
	} else if (strcmp(argv[i], "-N") == 0) {
	    if (++i < argc) {
		iterations = strtoul(argv[i], NULL, 10);
	    } else {
		print_usage(argv[0]);
		return 1;
	    }
	} else {
	    print_usage(argv[0]);
	    return 1;
	}
    }

    if (threaded_version) {
	if (quiet_version) {
	    c1 = new QTClient("Yanyan", 5000000,17);
	    c2 = new QTClient("Antonio", 5000000,31);
	    c3 = new QTClient("Cyrus", 5000000,1);
	    c4 = new QTClient("Alex", 5000000,8);
	} else {
	    c1 = new TClient("Yanyan", 5000000,17);
	    c2 = new TClient("Antonio", 5000000,31);
	    c3 = new TClient("Cyrus", 5000000,1);
	    c4 = new TClient("Alex", 5000000,8);
	}
    } else {
	if (quiet_version) {
	    c1 = new QClient("Yanyan", 5000000,17);
	    c2 = new QClient("Antonio", 5000000,31);
	    c3 = new QClient("Cyrus", 5000000,1);
	    c4 = new QClient("Alex", 5000000,8);
	} else {
	    c1 = new Client("Yanyan", 5000000,17);
	    c2 = new Client("Antonio", 5000000,31);
	    c3 = new Client("Cyrus", 5000000,1);
	    c4 = new Client("Alex", 5000000,8);
	}
    }

    Sim::create_process(c1);
    Sim::create_process(c2);
    Sim::create_process(c3);
    Sim::create_process(c4);

    Sim::run_simulation();

    delete(c1);
    delete(c2);
    delete(c3);
    delete(c4);
    return 0;
}


